"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConditionsBuilder = exports.OPERATORS = void 0;
const exceptions_1 = require("./exceptions");
const utils_1 = require("./utils");
exports.OPERATORS = {
    EQ: "=",
    NE: "!=",
    GT: ">",
    ST: "<",
    GE: ">=",
    SE: "<=",
    NG: "!>",
    NS: "!<",
    LK: "LIKE",
    NL: "NOT LIKE",
    RG: "REGEXP",
    IN: "IN",
    NI: "NOT IN",
    BT: "BETWEEN",
    NB: "NOT BETWEEN",
};
class ConditionsBuilder {
    constructor() {
        this.trimLeft = (string, charlist) => {
            if (charlist === undefined)
                charlist = " ";
            return string.replace(new RegExp("^[" + charlist + "]+"), "");
        };
        this.trimRight = (string, charlist) => {
            if (charlist === undefined)
                charlist = " ";
            return string.replace(new RegExp("[" + charlist + "]+$"), "");
        };
        this.trim = (string, charlist) => {
            return this.trimRight(this.trimLeft(string, charlist), charlist);
        };
        this.forEach = (Object, callback) => {
            const Results = [];
            for (let key in Object) {
                if (Object.hasOwnProperty(key)) {
                    const Result = callback(Object[key], key, Object);
                    if (typeof Result === "boolean") {
                        if (!Result)
                            break;
                        else
                            continue;
                    }
                    else
                        Results.push(Result);
                }
            }
            return Results;
        };
        this.processCondition = (conditions) => {
            if (typeof conditions !== "object" || conditions instanceof Array)
                throw new exceptions_1.SQLException(`Invalid conditions have been provided!`);
            const Query = {
                sql: "",
                values: [],
            };
            if (conditions instanceof utils_1.SQLMatch) {
                // Prepare Match Against Statement
                if (conditions.search) {
                    Query.sql =
                        [
                            `MATCH(\`${conditions.columns.join("`, `")}\`) AGAINST(?`,
                            conditions.modifier,
                            conditions.relevance,
                        ]
                            .filter((v) => v)
                            .join(" ") + `) ${conditions.operation || "AND"} `;
                    Query.values.push(conditions.search);
                }
                // Process Where Clause
                if (typeof conditions.WHERE === "object") {
                    let Where;
                    if (conditions.WHERE instanceof Array)
                        Where = this.processConditions(conditions.WHERE);
                    else
                        Where = this.processCondition(conditions.WHERE);
                    Query.values = Query.values.concat(Where.values);
                    if (Where.sql)
                        Query.sql += `(${this.trim(this.trim(Where.sql, " OR "), " AND ")})`;
                }
            }
            else {
                this.forEach(conditions, (value, key) => {
                    if (typeof value !== "undefined")
                        if (value instanceof Array) {
                            if (value.length) {
                                Query.sql += `\`${key}\` IN (${value.map(() => "?")}) AND `;
                                Query.values = Query.values.concat(value);
                            }
                        }
                        else if (typeof value === "object" && value !== null) {
                            // Collect Sub Queries
                            const SubQuery = {
                                sql: "",
                                values: [],
                            };
                            this.forEach(value, (v, operator) => {
                                if (typeof v !== "undefined") {
                                    if (!exports.OPERATORS[operator])
                                        throw new exceptions_1.SQLException(`Invalid Operator '${operator}' has been provided!`);
                                    else if (v instanceof Array) {
                                        // Remove Undefined Values
                                        v = v.filter((i) => typeof i !== "undefined");
                                        if (v.length)
                                            if (!["IN", "NI", "BT", "NB"].includes(operator))
                                                throw new exceptions_1.SQLException(`Value cannot be an Array with '${exports.OPERATORS[operator]}' Operator!`);
                                            else if (["IN", "NI"].includes(operator)) {
                                                SubQuery.sql += `\`${key}\` ${exports.OPERATORS[operator]} (${v.map((_) => "?")}) OR `;
                                                SubQuery.values = SubQuery.values.concat(v);
                                            }
                                            else {
                                                if (v.length < 2)
                                                    throw new exceptions_1.SQLException(`2 Values are required as Array in Between/Not Between conditions!`);
                                                SubQuery.sql += `(\`${key}\` ${exports.OPERATORS[operator]} ? AND ?) OR `;
                                                SubQuery.values = SubQuery.values.concat([v[0], v[1]]);
                                            }
                                    }
                                    else if (typeof v === "object" && v !== null)
                                        throw new exceptions_1.SQLException(`Condition value cannot be an Object!`);
                                    else if (["IN", "NI"].includes(operator)) {
                                        SubQuery.sql += `\`${key}\` ${exports.OPERATORS[operator]} (?) OR `;
                                        SubQuery.values.push(v);
                                    }
                                    else {
                                        SubQuery.sql += `\`${key}\` ${exports.OPERATORS[operator]} ? OR `;
                                        SubQuery.values.push(v);
                                    }
                                }
                            });
                            Query.sql += SubQuery.sql
                                ? `(${this.trim(SubQuery.sql, " OR ")}) AND `
                                : "";
                            Query.values = Query.values.concat(SubQuery.values);
                        }
                        else {
                            Query.sql += `\`${key}\` = ? AND `;
                            Query.values.push(value);
                        }
                });
            }
            // Trim Extra Operator
            if (Query.sql)
                Query.sql = Query.sql ? `(${this.trim(Query.sql, " AND ")}) OR ` : "";
            return Query;
        };
        this.processConditions = (conditions) => {
            if (conditions instanceof Array) {
                const Query = conditions
                    .map((condition) => this.processCondition(condition))
                    .reduce((p, c) => {
                    p.sql += c.sql;
                    p.values = p.values.concat(c.values);
                    return p;
                });
                // Trim Extra Operator
                Query.sql = this.trim(Query.sql, " OR ");
                return Query;
            }
            else
                throw new exceptions_1.SQLException(`Invalid conditions provided!`);
        };
        this.processGroupClause = (group) => {
            if (group.length)
                return `GROUP BY \`${group.join("`, `")}\` `;
            else
                return ``;
        };
        this.processOrderClause = (order) => {
            if (order.SORT && ["asc", "desc"].indexOf(order.SORT.toLowerCase()) < 0)
                order.SORT = "ASC";
            if (order.COLUMN)
                return `ORDER BY \`${order.COLUMN}\` ${order.SORT || "ASC"} `;
            else if (order.SORT)
                return `ORDER BY \`_id\` ${order.SORT} `;
            else
                return ``;
        };
        this.processLimitClause = (limit) => {
            if (limit.OFFSET && isNaN(limit.OFFSET))
                limit.OFFSET = 0;
            if (limit.NUMBER && isNaN(limit.NUMBER))
                delete limit.NUMBER;
            if (limit.NUMBER && limit.NUMBER >= 1)
                return `LIMIT ${limit.OFFSET || "0"} , ${limit.NUMBER || ""} `;
            else
                return ``;
        };
        this.processQueryConditions = (conditions, enableSelectFeatureClauses = true) => {
            const Query = {
                sql: "",
                values: [],
            };
            if (conditions.WHERE) {
                // Collect WHERE Clause SQL
                let Where;
                if (conditions.WHERE instanceof Array)
                    Where = this.processConditions(conditions.WHERE);
                else
                    Where = this.processCondition(conditions.WHERE);
                Query.values = Query.values.concat(Where.values);
                if (Where.sql)
                    Query.sql += `WHERE (${this.trim(Where.sql, " OR ")}) `;
            }
            // Select Feature Clauses cannot be used while updating or deleting record with conditions, So they should be disabled.
            if (enableSelectFeatureClauses) {
                // Collect GROUP BY Clause SQL
                if (conditions.GROUP)
                    Query.sql += this.processGroupClause(conditions.GROUP);
                // Collect HAVING Clause SQL
                if (conditions.HAVING) {
                    let Having;
                    if (conditions.HAVING instanceof Array)
                        Having = this.processConditions(conditions.HAVING);
                    else
                        Having = this.processCondition(conditions.HAVING);
                    Query.values = Query.values.concat(Having.values);
                    if (Having.sql)
                        Query.sql += `HAVING (${this.trim(Having.sql, " OR ")}) `;
                }
                // Collect ORDER BY Clause SQL
                if (conditions.ORDER)
                    Query.sql += this.processOrderClause(conditions.ORDER);
                // Collect LIMIT Clause SQL
                if (conditions.LIMIT)
                    Query.sql += this.processLimitClause(conditions.LIMIT);
            }
            // Trim Extra Operator
            Query.sql = Query.sql.trim();
            return Query;
        };
        this.conditions = (conditions) => {
            if (typeof conditions === "object") {
                // Condition Where
                this.WHERE =
                    typeof conditions.WHERE === "object"
                        ? conditions.WHERE
                        : undefined;
                // Extra Conditions
                this.HAVING =
                    typeof conditions.HAVING === "object"
                        ? conditions.HAVING
                        : undefined;
                this.GROUP =
                    typeof conditions.GROUP === "object"
                        ? conditions.GROUP
                        : undefined;
                this.ORDER =
                    typeof conditions.ORDER === "object"
                        ? conditions.ORDER
                        : undefined;
                this.LIMIT =
                    typeof conditions.LIMIT === "object"
                        ? conditions.LIMIT
                        : undefined;
            }
            // Return Builder Object
            return this;
        };
        this.reset = () => this.conditions({});
        this.render = (enableSelectFeatureClauses = true) => this.processQueryConditions(this, enableSelectFeatureClauses);
        this.toString = () => {
            const query = this.render();
            return query.sql
                .split("?")
                .map((queryPart, key) => queryPart + (query.values[key] ? `'${query.values[key]}'` : ""))
                .join("");
        };
    }
}
exports.ConditionsBuilder = ConditionsBuilder;
